#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <error.h>
#include <errno.h>

#include <unistd.h>
#include <fcntl.h>

#include <expat.h>

const char datafile[]="/usr/share/X11/xkb/rules/xorg.xml";
const char english[]="en_US";
const char *user_language;

typedef struct user_data_struct {
    char *el;
    char *lang;
    FILE *file;
    
    int mode;
    
    char *layname;
    char *name;
    
    char *laydescription;
    char *llaydescription;
    char *description;
    char *ldescription;
} user_data;

void 
do_free(char **item)
{
    if (*item) free(*item);
    *item = 0;
}

void 
do_assign(char **target,const char *destination,int len)
{
    do_free(target);
    
    if (destination && (len > 0))
    {
        *target=malloc(len+1);
        strncpy(*target,destination,len);
        (*target)[len]=0;
    }
}

const char *
get_lang(const char **attr)
{
    for (;*attr;attr+=2)
    {
	const char *name = *attr;
	const char *value = *(attr + 1);
	if (!strcmp(name,"xml:lang")) return value;
    }
    return english;
}

FILE *
xfopen(const char *name,const char *mode)
{
    FILE *f=fopen(name,mode);
    if (!f)
	error(EXIT_FAILURE,errno,"unable to open file %s for writing",name);
    return f;
}

void
do_reset(user_data *u)
{
    do_free(&u->description);
    do_free(&u->ldescription);
}

static void XMLCALL
element_start(void *data, const char *el, const char **attr)
{
    user_data *u = data;

    u->el = strdup(el);
    u->lang = strdup(get_lang(attr));

    if (!strcmp(el,"modelList"))
	u->file = xfopen("./models","w");
    else if (!strcmp(el,"layoutList"))
	u->file = xfopen("./layouts","w");
    else if (!strcmp(el,"optionList"))
	u->file = xfopen("./options","w");
    else if (!strcmp(el,"layout") || !strcmp(el,"group"))
	u->mode = 1;
    else if (!strcmp(el,"variant") || !strcmp(el,"option"))
	u->mode = 0;
}

static void XMLCALL
element_stop(void *data, const char *el)
{
    user_data *u = data;

    do_free(&u->el);
    do_free(&u->lang);

    if (!strcmp(el,"modelList") || !strcmp(el,"layoutList") || !strcmp(el,"optionList"))
	fclose(u->file);
    else if (!strcmp(el,"layout"))
    {
	fprintf(u->file,"%s\t%s\n",
	    u->layname,
	    u->llaydescription?u->llaydescription:u->laydescription);
	do_reset(u);
    }
    else if (!strcmp(el,"variant"))
    {
	u->mode = 1;
	fprintf(u->file,"%s(%s)\t%s - %s\n",
	    u->layname,
	    u->name,
	    u->llaydescription?u->llaydescription:u->laydescription,
	    u->ldescription?u->ldescription:u->description);
        do_reset(u);
    }
    else if (!strcmp(el,"option") && !strcmp(u->layname,"grp"))
    {
	u->mode = 1;
	fprintf(u->file,"%s\t%s\n",
	    u->name,
	    u->ldescription?u->ldescription:u->description);
        do_reset(u);
    }
    else if (!strcmp(el,"model"))
    {
	fprintf(u->file,"%s\t%s\n",
	    u->name,
	    u->ldescription?u->ldescription:u->description);
        do_reset(u);
    }
}

static void XMLCALL
cdata_handler(void *data,const char *buffer, int len)
{
    user_data *u = data;

    if (!u->el) return;

    if (!strcmp(u->el,"name"))
	do_assign(u->mode?&u->layname:&u->name,buffer,len);
    else if (!strcmp(u->el,"description") &&
             !strcmp(u->lang,english))
	do_assign(u->mode?&u->laydescription:&u->description,buffer,len);
    else if (!strcmp(u->el,"description") &&
             !strcmp(u->lang,user_language))
	do_assign(u->mode?&u->llaydescription:&u->ldescription,buffer,len);
    else if (!strcmp(u->el,"description") &&
             !strncmp(u->lang,user_language,2))
	do_assign(u->mode?&u->llaydescription:&u->ldescription,buffer,len);
}

int 
main(int argc,char **argv)
{
    int fd;
    char buffer[BUFSIZ];
    size_t len;
    XML_Parser parser;
    user_data *data;
    
    if (argc >= 2)
	user_language=argv[1];
    else
	user_language=english;

    if (!(parser = XML_ParserCreate(NULL)))
	error(EXIT_FAILURE,0,"couldn't allocate memory for parser");

    if (!(data = calloc(1,sizeof(user_data))))
	error(EXIT_FAILURE,errno,"couldn't allocate memory for user_data");

    if ((fd = open(datafile,O_RDONLY)) < 0)
	error(EXIT_FAILURE,errno,"unable to open database file %s",datafile);

    XML_SetElementHandler(parser, element_start, element_stop);
    XML_SetCharacterDataHandler(parser,cdata_handler);
    XML_SetUserData(parser, data);

    while ((len = read(fd,buffer,(sizeof buffer) - 1)) > 0)
    {
	if (XML_Parse(parser, buffer, len, XML_FALSE) == XML_STATUS_ERROR)
	{
	    fprintf(stderr,"Parse error at line %lu:%s",
		    XML_GetCurrentLineNumber(parser),
		    XML_ErrorString(XML_GetErrorCode(parser)));
	    return EXIT_FAILURE;
	}
    }
    XML_Parse(parser, buffer, 0, XML_TRUE);
    XML_ParserFree(parser);

    return EXIT_SUCCESS;
}
